import math
import numpy as np
from astroquery.ipac.ned import Ned
from astropy.wcs import WCS
from photutils.segmentation import detect_sources
from photutils.background import Background2D, MedianBackground, LocalBackground, MMMBackground
from photutils.detection import DAOStarFinder
from astropy.stats import sigma_clipped_stats
from photutils.psf import PSFPhotometry, MoffatPSF
from photutils.aperture import CircularAperture

from ..utils.utility import useful_functions

class Masking:
    def __init__(self):
        pass
    
    @classmethod
    def adapt_mask(cls, image_data, header, error_data, galaxy_name, observatory, band, image_set, manual):
        if ("max" not in image_set.psf[galaxy_name]) or (image_set.psf[galaxy_name]["max"] == -1):
            psf_list = useful_functions.extract_values_recursive(image_set.psf, galaxy_name)
            sig_t = np.max(psf_list)
            sig_3 = np.median(psf_list) + 3 * np.std(psf_list)
            
            sig_t = np.min([sig_t, sig_3])
            
            image_set.psf = (galaxy_name, "max", sig_t)
            
        fwhm = image_set.psf[galaxy_name]["max"]  # in "
        pixel_scale = np.abs(header.get("CD1_1", 1.1e-4)) * 3600
        fwhm = fwhm / pixel_scale  # in pixel
        
        mask_image, masked_image, _ = cls.make_mask(cls, image_data, header, galaxy_name, fwhm, manual)
        masked_err = np.where(mask_image, 999.0, error_data)
        
        image_set.update_data(masked_image, galaxy_name, observatory, band)
        image_set.update_error(masked_err, galaxy_name, observatory, band)
    
    
    def make_mask(self, image, header, galaxy, psf_fwhm, manual):
        mask, masked, sky = self.py2dmask(self, image, header, galaxy, psf_fwhm)
        
        if manual is not None:
            star_mask = self.manual_mask(self, image, header, psf_fwhm, manual)
            masked *= star_mask
            
        return mask, masked, sky
    
    
    def manual_mask(self, image, header, psf_fwhm, manual):
        coords = manual.get("coord", None)
        rads = manual.get("radius", None)
        
        if any(i is None for i in [coords, rads]):
            raise ValueError("Manual values are not input.")
        elif len(coords) != len(rads):
            raise ValueError("Length of coordinates and radii are not same.")
        
        wcs = WCS(header)
        mask = np.zeros_like(image)
        
        for coord, rad in zip(coords, rads):
            star_loc = wcs.all_world2pix(coord[0], coord[1], 0)
            star_mask = CircularAperture(positions=star_loc, r=rad).to_mask().to_image(shape=image.shape)
            star_mask = np.where(star_mask > 0, 0, 1)
            mask += star_mask
        
        return mask
    
    
    def py2dmask(self, image, header, galaxy, psf_fwhm):
        ra, dec = Ned.query_object(galaxy)['RA', 'DEC'][0]
        
        wcs = WCS(header)
        x, y = wcs.all_world2pix(ra, dec, 0)
        
        from astropy.stats import sigma_clipped_stats
        mean, median, std = sigma_clipped_stats(image, sigma=10.0, mask=np.where(image == 0, True, False))
        
        from photutils.segmentation import detect_sources
        segment_map = detect_sources(image, threshold=5 * std, npixels=30)
        print(x, y, int(segment_map.data[int(y), int(x)]))
        try:
            segment_map.remove_label(label=int(segment_map.data[int(y), int(x)]), relabel=True)
        except Exception as e:
            print(e)
            print("Un-removed")
            pass
        
        import matplotlib.pyplot as plt
        plt.imshow(segment_map, origin='lower', cmap=segment_map.cmap, interpolation='nearest')
        
        checksum = np.sum(np.where(segment_map.data != 0, 0, 1))
        if checksum != 0:
            masked_image = np.where(segment_map.data == 0, image, np.nan)
        else:
            masked_image = image
        
        return np.where(segment_map.data == 0, False, True), masked_image, np.full_like(masked_image, fill_value=median)
    

    def own_mask(self, image, header, galaxy, psf_fwhm):
        mean, median, std = sigma_clipped_stats(image, sigma=3.0)
        
        daofind = DAOStarFinder(fwhm=psf_fwhm, threshold=15.*std)
        sources = daofind(image - median)
        
        if sources is None:
            print('No sources found.')
            return image, image, image

        self.x0, self.y0, self.a, self.b, self.theta = useful_functions.get_galaxy_radius(image)

        dist = self.is_point_in_rotated_ellipse(self, sources['xcentroid'], sources['ycentroid'])

        good = dist > 1
        filtered_sources = sources[good]
        if not filtered_sources.indices:
            return image, image, image

        psf_model = MoffatPSF()
        psf_model.alpha.fixed = False
        psf_model.flux.fixed = False
        fit_shape =(int(psf_fwhm*2.0) * 2 + 1, int(psf_fwhm*2.0) * 2 + 1)
        bkgstat = MMMBackground()
        localbkg_estimator = LocalBackground(5, 10, bkgstat)
        
        psfphot = PSFPhotometry(psf_model, fit_shape)
        phot = psfphot(image - median, init_params=filtered_sources["xcentroid", "ycentroid", "flux"],
                       localbkg_estimator=localbkg_estimator)
        
        if phot is None:
            print('No PSF photometry found.')
            return image, image, image
        
        print(f"Phot: {phot}")
        
        resid = psfphot.make_residual_image(image - median)
        mask_image = np.where(image - median - resid > 0, 1, 0)
        masked_image = resid.copy()
        
        return mask_image, masked_image, np.full_like(masked_image, fill_value=median)


    def __make_mask(self, image, header, galaxy, psf_fwhm):
        psf_fwhm = np.abs(psf_fwhm)
        ra, dec = Ned.query_object(galaxy)['RA', 'DEC'][0]
        image_x, image_y = image.shape
        box_shape = [int(image_x/5), int(image_y/5)]
        filter_shape = [int(psf_fwhm*2.0) * 2 + 1, int(psf_fwhm*2.0) * 2 + 1]
        
        wcs = WCS(header)
        x, y = wcs.all_world2pix(ra, dec, 0)

        bkg_estimator = MedianBackground()
        try:    
            bkg = Background2D(image, box_shape, filter_size=filter_shape, bkg_estimator=bkg_estimator)
        except ValueError:
            print('ValueError occured. Try Smaller Background size.')
            bkg = Background2D(image, (int(box_shape[0]/5), int(box_shape[1]/5)), filter_size=filter_shape, bkg_estimator=bkg_estimator)
        threshold = 1.5*bkg.background_rms

        segment_map = detect_sources(image, threshold, npixels=5)
        if segment_map == None:
            return image, image, image

        sky_map = np.nonzero(segment_map.data)
        sky_image = image.copy()
        sky_image[sky_map] = np.nan

        label_main = segment_map.data[int(y), int(x)]
        if label_main != 0:
            segment_map.remove_labels([label_main])

        mask = np.nonzero(segment_map.data)
        masked_image = image.copy()
        masked_image[mask] = np.nan
            
        mask_image = np.zeros_like(image)
        mask_image[mask] = image[mask]
        
        mean, median, std = sigma_clipped_stats(image, sigma=3.0)
        daofind = DAOStarFinder(fwhm=psf_fwhm, threshold=15.*std)
        sources = daofind(masked_image - mean)
        
        if sources is None:
            print('No sources found.')
            return masked_image, masked_image, masked_image

        self.x0, self.y0, self.a, self.b, self.theta = useful_functions.get_galaxy_radius(image)

        dist = self.is_point_in_rotated_ellipse(self, sources['xcentroid'], sources['ycentroid'])

        good = dist > 1
        filtered_sources = sources[good]
        if not filtered_sources.indices:
            return image, image, image

        psf_model = MoffatPSF()
        psf_model.alpha.fixed = False
        psf_model.flux.fixed = False
        fit_shape = (31, 31) 
        psfphot = PSFPhotometry(psf_model, fit_shape,
                                aperture_radius=psf_fwhm*5)
        phot = psfphot(masked_image - mean, init_params=filtered_sources["xcentroid", "ycentroid", "flux"])
        if phot is None:
            print('No PSF photometry found.')
            return masked_image, masked_image, masked_image
        
        resid = psfphot.make_residual_image(masked_image - mean)
        mask_image = np.where(masked_image - mean - resid > 0, 1, 0)
        masked_image = resid.copy()
        
        return mask_image, masked_image, sky_image
        

    def is_point_in_rotated_ellipse(self, x, y):

        dx = x - self.x0
        dy = y - self.y0

        cos_t = math.cos(self.theta)
        sin_t = math.sin(self.theta)
        x_rot =  dx * cos_t + dy * sin_t
        y_rot = -dx * sin_t + dy * cos_t

        value = (x_rot / self.a)**2 + (y_rot / self.b)**2
        return value



import numpy as np
from pathlib import Path
from astropy.io import fits
import pandas as pd
import re
import inspect
import warnings
from glob import glob
from functools import wraps
from typing import Callable, Optional, Union, List

from .utility import Filters, Observatories
from .utility import useful_functions
from .file_generator import *
from ..plot.plot import DrawGalaxy


class GalaxyImageSet:
    __signature__ = inspect.Signature()

    def __init__(self):
        self._data = {}
        self._header = {}
        self._error = {}
        self._psf = {}
        self._cutout_shape = {}
        self._obs = {}
        self._files = []
        
        
    def add_image(self, filepath):
        filepath = Path(filepath)
        if not filepath.exists():
            raise FileNotFoundError(f"File {filepath} not found")

        with fits.open(filepath) as hdul:
            image_data = hdul[0].data
            image_header = hdul[0].header

        file_name = filepath.stem
        galaxy_name = Parsers._galaxy_name_parser(file_name=file_name)
        band = Parsers._band_name_parser(file_name=file_name)
        observatory = Parsers._observatory_name_parser(file_name=file_name)
        if galaxy_name is None or band is None or observatory is None:
            return

        for attr in [self._data, self._header, self._error, self._psf, self._cutout_shape]:
            attr.setdefault(galaxy_name, {}).setdefault(observatory, {})

        self._data[galaxy_name][observatory][band] = image_data
        self._header[galaxy_name][observatory][band] = image_header
        self._psf[galaxy_name][observatory][band] = -1.0
        self._cutout_shape[galaxy_name][observatory][band] = set()
        self._files.append(str(filepath))

        err_name = filepath.with_name(filepath.stem + "_err").with_suffix(filepath.suffix)
        if not err_name.exists():
            warnings.warn(f"Error File {err_name} not found. Continue without error map.")
            self._error[galaxy_name][observatory][band] = np.zeros_like(image_data)
        else:
            with fits.open(err_name) as hdul_err:
                error_data = hdul_err[0].data
            self._error[galaxy_name][observatory][band] = error_data

    def update_data(self, image_data, galaxy_name, observatory, band):
        if not all([galaxy_name, observatory, band]):
            raise KeyError("Specify galaxy, observatory, and band.")
        try:
            self._data[galaxy_name][observatory][band] = image_data
        except KeyError:
            raise KeyError("Invalid galaxy/observatory/band specified")

    def update_error(self, error_data, galaxy_name, observatory, band):
        if not all([galaxy_name, observatory, band]):
            raise KeyError("Specify galaxy, observatory, and band.")
        try:
            self._error[galaxy_name][observatory][band] = error_data
        except KeyError:
            raise KeyError("Invalid galaxy/observatory/band specified")
    
        
    # merging / append instance
    def append(self, other):
        """
        Append data from another GalaxyImageSet instance to this instance.
        This modifies the current instance.
        
        Args:
            other (GalaxyImageSet): Another GalaxyImageSet instance to append from
        """
        if not isinstance(other, GalaxyImageSet):
            raise TypeError("Can only append another GalaxyImageSet instance")
        
        # Merge data
        for galaxy_name, observatories in other._data.items():
            if galaxy_name not in self._data:
                self._data[galaxy_name] = {}
            
            for observatory, bands in observatories.items():
                if observatory not in self._data[galaxy_name]:
                    self._data[galaxy_name][observatory] = {}
                
                for band, image_data in bands.items():
                    # Override if exists, add if new
                    self._data[galaxy_name][observatory][band] = image_data
        
        # Merge headers
        for galaxy_name, observatories in other._header.items():
            if galaxy_name not in self.header:
                self._header[galaxy_name] = {}
            
            for observatory, bands in observatories.items():
                if observatory not in self._header[galaxy_name]:
                    self._header[galaxy_name][observatory] = {}
                
                for band, header_data in bands.items():
                    self._header[galaxy_name][observatory][band] = header_data
        
        # Merge error data
        for galaxy_name, observatories in other.error.items():
            if galaxy_name not in self.error:
                self._error[galaxy_name] = {}
            
            for observatory, bands in observatories.items():
                if observatory not in self._error[galaxy_name]:
                    self._error[galaxy_name][observatory] = {}
                
                for band, error_data in bands.items():
                    self._error[galaxy_name][observatory][band] = error_data
        
        # Merge PSF data
        for galaxy_name, observatories in other._psf.items():
            if galaxy_name not in self._psf:
                self._psf[galaxy_name] = {}
            
            for observatory, bands in observatories.items():
                if observatory not in self._psf[galaxy_name]:
                    self._psf[galaxy_name][observatory] = {}
                
                for band, psf_data in bands.items():
                    self._psf[galaxy_name][observatory][band] = psf_data
        
        # Merge obs data
        for galaxy_name, obs_data in other._obs.items():
            if galaxy_name not in self.obs:
                self._obs[galaxy_name] = obs_data
            else:
                # If both have obs data for the same galaxy, merge them
                if isinstance(self._obs[galaxy_name], dict) and isinstance(obs_data, dict):
                    self._obs[galaxy_name].update(obs_data)
                else:
                    self._obs[galaxy_name] = obs_data
        
        # Merge file lists (avoid duplicates)
        for filepath in other._files:
            if filepath not in self._files:
                self._files.append(filepath)


    def merge(self, other):
        """
        Create a new GalaxyImageSet instance containing data from both instances.
        This does not modify either original instance.
        
        Args:
            other (GalaxyImageSet): Another GalaxyImageSet instance to merge with
            
        Returns:
            GalaxyImageSet: A new instance containing merged data
        """
        if not isinstance(other, GalaxyImageSet):
            raise TypeError("Can only merge with another GalaxyImageSet instance")
        
        # Create new instance
        merged = GalaxyImageSet()
        
        # First append self to the new instance
        merged.append(self)
        
        # Then append other to the new instance
        merged.append(other)
        
        return merged
    

    @property
    def files(self):
        """Read Only: Path of Files"""
        return tuple(self._files)
    
    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value_tuple):
        galaxy, observatory, band, image_data = value_tuple
        try:
            self._data[galaxy][observatory][band] = image_data
        except KeyError:
            raise KeyError("Invalid galaxy/observatory/band for setting image data.")
        
    @property
    def header(self):
        return self._header

    @header.setter
    def header(self, value_tuple):
        galaxy, observatory, band, header_data = value_tuple
        try:
            self._header[galaxy][observatory][band] = header_data
        except KeyError:
            raise KeyError("Invalid galaxy/observatory/band for setting image data.")
        
    @property
    def error(self):
        return self._error

    @error.setter
    def error(self, value_tuple):
        galaxy, observatory, band, image_data = value_tuple
        try:
            self._error[galaxy][observatory][band] = image_data
        except KeyError:
            raise KeyError("Invalid galaxy/observatory/band for setting image data.")

    @property
    def psf(self):
        return self._psf
    
    @psf.setter
    def psf(self, value_tuple):
        if not isinstance(value_tuple, tuple):
            raise TypeError("psf.setter expects a tuple.")
        
        if len(value_tuple) == 4:
            galaxy, observatory, band, new_val = value_tuple
            if galaxy not in self._psf:
                self._psf[galaxy] = {}
            if observatory not in self._psf[galaxy]:
                self._psf[galaxy][observatory] = {}
            self._psf[galaxy][observatory][band] = new_val

        elif len(value_tuple) == 3:
            galaxy, val_name, new_val = value_tuple
            if galaxy not in self._psf:
                self._psf[galaxy] = {}
            self._psf[galaxy][val_name] = new_val

        else:
            raise ValueError("value_tuple must be length 3 or 4.")

    @property
    def cutout_shape(self):
        return self._cutout_shape

    @cutout_shape.setter
    def cutout_shape(self, value_tuple):
        if not hasattr(value_tuple, '__iter__'):
            raise TypeError("cutout_shape setter expects an iterable (tuple or list)")
        if len(value_tuple) != 4:
            raise ValueError("cutout_shape setter expects length-4 tuple: (galaxy, observatory, band, box_shape)")

        galaxy, observatory, band, box_shape = value_tuple

        if not (isinstance(box_shape, (list, tuple))):
            raise ValueError("box_shape must be tuple/list")

        if galaxy not in self._cutout_shape:
            self._cutout_shape[galaxy] = {}
        if observatory not in self._cutout_shape[galaxy]:
            self._cutout_shape[galaxy][observatory] = {}

        self._cutout_shape[galaxy][observatory][band] = box_shape
        
        print(f"Set cutout_shape[{galaxy}][{observatory}][{band}] = {box_shape}")
        
    def summary(self):
        """Print summary of galaxies and available bands."""
        for galaxy, val in self._data.items():
            for obs, bands in val.items():
                galaxy = sorted(galaxy) if isinstance(galaxy, list) else [galaxy]
                obs = sorted(obs) if isinstance(obs, list) else [obs]
                print(f"Galaxy: {galaxy}\nObservatories: {obs} \nBands: {sorted(bands.keys())}")
                
    # ==== Plotting Properties ====
    def plot_image(self, value_tuple):
        if not hasattr(value_tuple, '__iter__'):
            raise TypeError("plot_image expects an iterable (tuple or list)")
        if len(value_tuple) != 3:
            raise ValueError("plot_image expects length-3 tuple: (galaxy, observatory, band)")
        
        galaxy, observatory, band = value_tuple
        try:
            fig, ax = DrawGalaxy.single_galaxy(self, galaxy, observatory, band)
            return fig, ax
        except KeyError:
            raise KeyError("Invalid galaxy/observatory/band for plotting image data.")
        

class Parsers:
    def __init__(self):
        pass
    
    @staticmethod
    def _observatory_name_parser(file_name):
        """Parse observatory names from a given string or list."""
        if not isinstance(file_name, str):
            raise ValueError("file_name must be a string")
        for obs in Observatories.get_observatories():
            pattern = "|".join(map(re.escape, ['-', ' ', '_', '.']))
            if obs in re.split(pattern, file_name):
                return obs
        return None
    
    @staticmethod
    def _galaxy_name_parser(file_name):
        """Parse galaxy names from a given string or list."""
        if not isinstance(file_name, str):
            raise ValueError("file_name must be a string")
        for galaxy_category in ['NGC', 'IC', 'M']:
            pattern = "|".join(map(re.escape, ['-', ' ', '_', '.']))
            match = [g for g in re.split(pattern, file_name) if galaxy_category in g]
            if match:
                return match[0]
        return None
    
    @staticmethod
    def _band_name_parser(file_name):
        """Parse band names from a given string or list."""
        if not isinstance(file_name, str):
            raise ValueError("file_name must be a string")
        for band in Filters.get_all_filters():
            pattern = "|".join(map(re.escape, ['-', ' ', '_', '.']))
            if band in re.split(pattern, file_name):
                return band
        return None
    

class ImageQuery():
    def __init__(self):
        pass    
                
    def queryImages(self, dir, galaxy_name=None, band=None):
        """
        Query images from a directory based on galaxy names and bands.
        Returns a GalaxyImageSet object containing the queried images.
        """
        image_set = GalaxyImageSet()
        dir_path = Path(dir).parent
        file_pattern = Path(dir).name if Path(dir).name.endswith('.fits') else "*.fits"
        
        if not dir_path.is_dir():
            raise NotADirectoryError(f"{dir} is not a valid directory")

        if isinstance(galaxy_name, str):
            galaxy_name = [galaxy_name]
        if isinstance(band, str):
            band = [band]

        for fits_file in dir_path.glob(file_pattern):
            try:
                image_set.add_image(fits_file)
            except Exception as e:
                print(f"Error processing {fits_file}: {e}")
                    
        return image_set
    
    @classmethod
    def queryAllImages(cls, dir):
        image_set = GalaxyImageSet()
        
        if not isinstance(dir, str):
            dir = str(dir)
        
        for file in list(glob(dir)):
            if Path(file).stem.endswith("_err"):
                continue
            image_set.add_image(file)
            
        return image_set


import numpy as np
import math
import inspect
import shutil
from pathlib import Path
from typing import Union, Tuple, Dict, List, Optional
from dataclasses import dataclass
from importlib import resources
from astropy.coordinates import SkyCoord
import astropy.units as u
from photutils.segmentation import detect_threshold, detect_sources, SourceCatalog


@dataclass
class FilterCurve:
    """Container for filter response curve data."""
    name: str
    wavelength: np.ndarray
    response: np.ndarray
    source_type: str  # 'default', 'file', 'array'
    source_path: Optional[str] = None
    unit_type: str = 'photon'  # 'photon' or 'energy'
    description: str = ''
    
    def __post_init__(self):
        """Validate and convert arrays to numpy."""
        self.wavelength = np.asarray(self.wavelength)
        self.response = np.asarray(self.response)
        
        if len(self.wavelength) != len(self.response):
            raise ValueError("Wavelength and response arrays must have same length")
    
    def save_to_file(self, filepath: Union[str, Path]):
        """Save filter curve to ASCII .dat file with proper header format."""
        filepath = Path(filepath)
        
        # Create header lines
        header_lines = [
            f"# {self.name}",
            f"# {self.unit_type}",
            f"# {self.description}"
        ]
        
        # Write file manually to control header format exactly
        with open(filepath, 'w') as f:
            # Write header
            for line in header_lines:
                f.write(line + '\n')
            
            # Write data
            for wl, resp in zip(self.wavelength, self.response):
                f.write(f"{wl:.3f} {resp:.3f}\n")

class Filters:
    """Class to handle different photometric filters and their properties."""
    
    # Default filter names (curves loaded lazily)
    _default_broadband = [
        'FUV', 'NUV', 'u', 'g', 'r', 'i', 'z', 'Y', 'J', 'H', 'Ks',
        'w1', 'w2', 'w3', 'w4', "ch1", "ch2", "ch3", "ch4", "24mu", "70mu", "160mu"
    ]
    _default_mediumband = [f'm{wave}' for wave in range(400, 900, 25)]
    
    # Storage for loaded filter curves
    _loaded_curves: Dict[str, FilterCurve] = {}
    _custom_filters: Dict[str, FilterCurve] = {}
    
    def __init__(self):
        """Initialize filter instance."""
        self.broadband = self._default_broadband.copy()
        self.mediumband = self._default_mediumband.copy()
        self.filters = self.broadband + self.mediumband + list(self._custom_filters.keys())
    
    @classmethod
    def _load_default_filter(cls, filter_name: str) -> FilterCurve:
        """Load a default filter from package data."""
        if filter_name in cls._loaded_curves:
            return cls._loaded_curves[filter_name]
        
        try:
            # Get the filter curves directory
            filter_dir = resources.files("Spec7DT.reference.filter_curves")
            
            # Search for files that contain the filter name
            matching_files = []
            with resources.as_file(filter_dir) as dir_path:
                if dir_path.is_dir():
                    for file_path in dir_path.glob("*.dat"):
                        # Check if filter_name is in the filename
                        if f".{filter_name}." in file_path.name or file_path.stem.endswith(f".{filter_name}"):
                            matching_files.append(file_path)
                    
                    # If no exact match, try looser matching
                    if not matching_files:
                        for file_path in dir_path.glob("*.dat"):
                            if filter_name in file_path.name:
                                matching_files.append(file_path)
            
            if not matching_files:
                raise FileNotFoundError(f"No filter file found containing '{filter_name}'")
            
            if len(matching_files) > 1:
                file_names = [f.name for f in matching_files]
                print(f"Warning: Multiple files found for '{filter_name}': {file_names}")
                print(f"Using: {matching_files[0].name}")
            
            # Use the first matching file
            selected_file = matching_files[0]
            
            # Load data and header information
            wavelength, response, file_filter_name, unit_type, description = cls._load_file_filter(selected_file)
                
            curve = FilterCurve(
                name=filter_name,  # Use requested name, not file header name
                wavelength=wavelength,
                response=response,
                source_type='default',
                source_path=str(selected_file),
                unit_type=unit_type,
                description=description
            )
            
            cls._loaded_curves[filter_name] = curve
            return curve
            
        except Exception as e:
            raise FileNotFoundError(f"Could not load default filter '{filter_name}': {e}")
    
    @classmethod
    def _load_file_filter(cls, filepath: Union[str, Path]) -> Tuple[np.ndarray, np.ndarray, str, str, str]:
        """Load filter data from ASCII .dat file and parse header."""
        filepath = Path(filepath)
        if not filepath.exists():
            raise FileNotFoundError(f"Filter file not found: {filepath}")
        
        try:
            # Read header information
            filter_name = ""
            unit_type = "photon"  # default
            description = ""
            
            with open(filepath, 'r') as f:
                lines = f.readlines()
                
            # Parse first three header lines
            header_count = 0
            
            for i, line in enumerate(lines):
                if line.startswith('#'):
                    content = line[1:].strip()  # Remove # and whitespace
                    if header_count == 0:
                        filter_name = content
                    elif header_count == 1:
                        unit_type = content if content in ['photon', 'energy'] else 'photon'
                    elif header_count == 2:
                        description = content
                    header_count += 1
                else:
                    break
            
            # Use filename stem if no filter name in header
            if not filter_name:
                filter_name = filepath.stem
            
            # Load numerical data
            data = np.loadtxt(filepath)
            if data.ndim != 2 or data.shape[1] != 2:
                raise ValueError("File must contain 2 columns: wavelength and response")
                
            return data[:, 0], data[:, 1], filter_name, unit_type, description
            
        except Exception as e:
            raise ValueError(f"Could not read filter file '{filepath}': {e}")
    
    @classmethod
    def add_filter_from_file(cls, filepath: Union[str, Path], filter_name: Optional[str] = None, verbose=False):
        """Add custom filter from ASCII .dat file."""
        filepath = Path(filepath)
        
        wavelength, response, file_filter_name, unit_type, description = cls._load_file_filter(filepath)
        
        # Use provided name, or file header name, or filename as fallback
        final_name = filter_name or file_filter_name or filepath.stem
        
        curve = FilterCurve(
            name=final_name,
            wavelength=wavelength,
            response=response,
            source_type='file',
            source_path=str(filepath.resolve()),
            unit_type=unit_type,
            description=description
        )
        
        cls._custom_filters[final_name] = curve
        if verbose:
            print(f"Added filter '{final_name}' from file: {filepath}")
            if description:
                print(f"  Description: {description}")
                print(f"  Unit type: {unit_type}")
    
    @classmethod
    def add_filter_from_arrays(cls, wavelength: Union[List, np.ndarray], 
                              response: Union[List, np.ndarray], filter_name: str,
                              unit_type: str = 'photon', description: str = ''):
        """Add custom filter from numpy arrays or lists."""
        curve = FilterCurve(
            name=filter_name,
            wavelength=wavelength,
            response=response,
            source_type='array',
            unit_type=unit_type,
            description=description
        )
        
        cls._custom_filters[filter_name] = curve
        print(f"Added filter '{filter_name}' from arrays")
    
    @classmethod
    def remove_filter(cls, filter_name: str) -> bool:
        """Remove a custom filter."""
        if filter_name in cls._custom_filters:
            del cls._custom_filters[filter_name]
            print(f"Removed custom filter: {filter_name}")
            return True
        else:
            print(f"Filter '{filter_name}' not found in custom filters")
            return False
    
    @classmethod
    def get_filter_curve(cls, filter_name: str, observatory: Optional[str] = None) -> FilterCurve:
        """Get filter response curve."""
        # Create full filter key for caching
        full_key = f"{observatory}.{filter_name}" if observatory else filter_name
        
        # Check custom filters first
        if filter_name in cls._custom_filters:
            return cls._custom_filters[filter_name]
        
        # Check if it's a default filter
        if filter_name in cls._default_broadband + cls._default_mediumband:
            # If observatory specified, try to load specific observatory filter
            if observatory:
                return cls._load_default_filter_with_observatory(filter_name, observatory)
            else:
                return cls._load_default_filter(filter_name)
        
        raise ValueError(f"Filter '{filter_name}' not found")
    
    @classmethod
    def _load_default_filter_with_observatory(cls, filter_name: str, observatory: str) -> FilterCurve:
        """Load a specific observatory filter."""
        cache_key = f"{observatory}.{filter_name}"
        
        if cache_key in cls._loaded_curves:
            return cls._loaded_curves[cache_key]
        
        try:
            # Look specifically for observatory.filter.dat
            filter_dir = resources.files("Spec7DT.reference.filter_curves")
            target_filename = f"{observatory}.{filter_name}.dat"
            
            with resources.as_file(filter_dir / target_filename) as dat_file:
                if dat_file.exists():
                    wavelength, response, file_filter_name, unit_type, description = cls._load_file_filter(dat_file)
                    
                    curve = FilterCurve(
                        name=f"{observatory}.{filter_name}",
                        wavelength=wavelength,
                        response=response,
                        source_type='default',
                        source_path=str(dat_file),
                        unit_type=unit_type,
                        description=description
                    )
                    
                    cls._loaded_curves[cache_key] = curve
                    return curve
                else:
                    raise FileNotFoundError(f"Specific filter file not found: {target_filename}")
            
        except Exception as e:
            # Fall back to general filter search
            print(f"Could not find {observatory}.{filter_name}.dat, trying general search...")
            return cls._load_default_filter(filter_name)
    
    @classmethod
    def get_all_filters(cls) -> List[str]:
        """Return list of all available filters."""
        return cls._default_broadband + cls._default_mediumband + list(cls._custom_filters.keys())
    
    @classmethod
    def get_custom_filters(cls) -> List[str]:
        """Return list of custom filters only."""
        return list(cls._custom_filters.keys())
    
    @classmethod
    def interpolate_filter(cls, filter_name: str, new_wavelength: np.ndarray) -> np.ndarray:
        """Interpolate filter response to new wavelength grid."""
        curve = cls.get_filter_curve(filter_name)
        return np.interp(new_wavelength, curve.wavelength, curve.response, left=0, right=0)
    
    @classmethod
    def save_custom_filter(cls, filter_name: str, filepath: Union[str, Path]):
        """Save a custom filter to file."""
        if filter_name not in cls._custom_filters:
            raise ValueError(f"Custom filter '{filter_name}' not found")
        
        curve = cls._custom_filters[filter_name]
        curve.save_to_file(filepath)
        print(f"Saved filter '{filter_name}' to: {filepath}")
    
    @classmethod
    def list_filter_info(cls):
        """Print information about all filters."""
        print(f"Default broadband filters ({len(cls._default_broadband)}): {', '.join(cls._default_broadband)}")
        print(f"Default mediumband filters ({len(cls._default_mediumband)}): {len(cls._default_mediumband)} filters")
        print(f"Custom filters ({len(cls._custom_filters)}):")
        
        for name, curve in cls._custom_filters.items():
            wl_range = f"{curve.wavelength.min():.0f}-{curve.wavelength.max():.0f} Å"
            print(f"  {name}: {len(curve.wavelength)} points, {wl_range}, source: {curve.source_type}")
    
    @classmethod
    def clear_custom_filters(cls):
        """Remove all custom filters."""
        count = len(cls._custom_filters)
        cls._custom_filters.clear()
        print(f"Cleared {count} custom filters")
    
    @classmethod
    def get_catcols(cls, cat_type, col_names):
        """Return a dictionary of given type"""
        catcols ={"cigale": cls.cigale,
                  "eazy": cls.eazy,
                  "lephare": cls.lephare,
                  "ppxf": cls.ppxf,
                  "goyangyi": cls.goyangyi
                  }
        
        function = catcols[cat_type.lower()]
        sig = inspect.signature(function)
        
        # image_data, header, error_data, galaxy_name, observatory, band, image_set
        kwargs = {"self": cls, "col_names": col_names}
        
        filtered_kwargs = {k: v for k, v in kwargs.items() if k in sig.parameters}
        return function(**filtered_kwargs)
    
    
    def cigale(self):
        cols_cigale = {
            'GALEX.NUV': 'galex.NUV',
            'GALEX.FUV': 'galex.FUV',
            'SDSS.u': 'sloan.sdss.u',
            'SDSS.g': 'sloan.sdss.g',
            'SDSS.r': 'sloan.sdss.r',
            'SDSS.i': 'sloan.sdss.i',
            'SDSS.z': 'sloan.sdss.z',
            'PanStarrs.y': 'PAN-STARRS_y',
            '2MASS.J': 'J_2mass',
            '2MASS.H': 'H_2mass',
            '2MASS.Ks': 'Ks_2mass',
            'Spitzer.ch1': 'spitzer.irac.ch1',
            'Spitzer.ch2': 'spitzer.irac.ch2',
            'Spitzer.ch3': 'spitzer.irac.ch3',
            'Spitzer.ch4': 'spitzer.irac.ch4',
            'WISE.w1': 'wise.W1',
            'WISE.w2': 'wise.W2',
            'WISE.w3': 'wise.W3',
            'WISE.w4': 'wise.W4',
            'F657N': 'HST.UVIS1.F657N',
            'F658N': 'HST.UVIS1.F658N',
        }
        cols_cigale.update({f'{key}_err': f'{cols_cigale[key]}_err' for key in cols_cigale.keys() if '_err' not in key})
        return cols_cigale
    
    def eazy(self, col_names):
        flux_dict = {name:f"F_{name}" for name in col_names if "_err" not in name}
        err_dict = {name:f"E_{name.strip('_err')}" for name in col_names if "_err" in name}
        flux_dict.update(err_dict)
        
        cols_eazy = flux_dict
        return cols_eazy
    
    def lephare(self):
        cols_lephare = {
            
        }
        return cols_lephare
    
    def ppxf(self):
        cols_ppxf = {
            
        }
        return cols_ppxf
    
    
    def goyangyi(self):
        cols_cigale = self.cigale(self)
        print(" ╱|、\n(˚ˎ 。7  \n |、˜〵          \n じしˍ,)ノ")
        return cols_cigale

class Observatories:
    """Class to handle different observatories and their properties."""
    def __init__(self):
        self.optical_obs = self._opticals()
        self.ir_obs = self._infrareds()
        self.uv_obs = self._ultraviolet()
        self.radio_obs = self._radio()
        self.observatories = list(set(self.optical_obs + self.ir_obs + self.uv_obs + self.radio_obs))
        
    def _opticals(self):
        """Return a list of optical observatories."""
        return ['HST', 'SDSS', 'PS1', 'CFHT', 'DECam', 'DES', 'LSST', 'Pan-STARRS', 'Subaru', '7DT', 'SkyMapper']
    
    def _infrareds(self):
        """Return a list of infrared observatories."""
        return ['WISE', 'Spitzer', 'Herschel', 'JWST', 'VISTA', 'UKIDSS', '2MASS', 'SPHEREx']
    
    def _ultraviolet(self):
        """Return a list of ultraviolet observatories."""
        return ['GALEX', 'HST', 'FUSE']
    
    def _radio(self):
        """Return a list of radio observatories."""
        return ['VLA', 'ALMA', 'LOFAR', 'SKA', 'MeerKAT', 'GMRT']
    
    @classmethod
    def get_observatories(cls):
        """Return a list of all observatories."""
        return cls().observatories
    

class useful_functions:
    @classmethod
    def get_redshift(cls, galaxy_name):
        """
        Query redshift from NED using galaxy name.
        
        Parameters:
        -----------
        galaxy_name : str
            Name of the galaxy (e.g., 'NGC 3627', 'M81', 'NGC4321')
        
        Returns:
        --------
        float or None
            Redshift value, or None if not found
        """
        from astroquery.ned import Ned
        
        try:
            # Query basic information from NED
            result_table = Ned.query_object(galaxy_name)
            
            if len(result_table) > 0:
                # Get the redshift value
                redshift = result_table['Redshift'][0]
                
                # Check if redshift is valid (not masked or NaN)
                if not np.ma.is_masked(redshift) and not np.isnan(redshift):
                    return float(redshift)
                else:
                    print(f"No redshift data available for {galaxy_name}")
                    return None
            else:
                print(f"Galaxy {galaxy_name} not found in NED")
                return None
                
        except Exception as e:
            print(f"Error querying {galaxy_name}: {str(e)}")
            return None
    
    @classmethod
    def get_galaxy_radius(cls, image):
        threshold = detect_threshold(image, nsigma=3.0)

        segm = detect_sources(image, threshold, npixels=5)
        if segm is None:
            print("No sources detected.")
            a, b = image.shape
            x0, y0 = image.shape[0]/2, image.shape[1]/2
            theta = 0
            return x0, y0, a, b, theta

        catalog = SourceCatalog(image, segm)
        gal = max(catalog, key=lambda src: src.area)

        x0, y0 = gal.xcentroid, gal.ycentroid
        a, b = gal.semimajor_sigma.value*2, gal.semiminor_sigma.value*2
        theta = math.radians(gal.orientation.value)
        return x0, y0, a, b, theta
    
    @staticmethod
    def find_rec(N):
        # Start from the square root of N and work downwards
        num_found = False
        while not num_found:
            for k in range(int(N ** 0.5), 0, -1):
                if N % k == 0:  # k must divide N
                    l = N // k  # Calculate l
                    # Check the condition that neither exceeds twice the other
                    if k <= 2 * l and l <= 2 * k:
                        num_found = True
                        return k, l
            N = N + 1
        return None, None  # Return None if no valid pair is found
    
    @staticmethod
    def get_pixel_scale(header, typical=1.0):
        cd1_1 = np.abs(header.get("CD1_1", header.get("PC1_1", typical)))
        cd1_2 = np.abs(header.get("CD1_2", header.get("PC1_2", 0.00)))
        pixel_scale = 3600 * np.sqrt(cd1_1 ** 2 + cd1_2 ** 2)
        return pixel_scale
    
    @staticmethod
    def extract_values_recursive(dictionary, key):
        """
        Alternative recursive approach that handles arbitrary nesting depth.
        
        Args:
            dictionary: Dictionary with nested structure
            key: The key at level1 to extract values from
        
        Returns:
            List of all values found in the nested structure
        """
        def _extract_all_values(obj):
            """Recursively extract all values from nested dict/list structures."""
            if isinstance(obj, dict):
                values = []
                for v in obj.values():
                    values.extend(_extract_all_values(v))
                return values
            elif isinstance(obj, list):
                values = []
                for item in obj:
                    values.extend(_extract_all_values(item))
                return values
            else:
                return [obj]
        
        if key not in dictionary:
            return []
        
        return _extract_all_values(dictionary[key])
    
    @classmethod
    def tour_nested_dict_with_keys(cls, dictionary):
        """
        Tour through a 3-level nested dictionary and yield keys and values in order.
        
        Args:
            dictionary: Dictionary with structure dict[level1][level2][level3]
        
        Yields:
            tuple: (keys_tuple, value) where keys_tuple contains (level1_key, level2_key, level3_key)
        """
        for level1_key, level1_dict in dictionary.items():
            for level2_key, level2_dict in level1_dict.items():
                for level3_key, value in level2_dict.items():
                    yield (level1_key, level2_key, level3_key), value


    def get_all_keys_and_values(self, my_dict):
        """
        Get all keys and values from a 3-level nested dictionary as a list.
        
        Args:
            my_dict: Dictionary with structure dict[level1][level2][level3]
        
        Returns:
            list: List of tuples [(keys_tuple, value), ...] where keys_tuple contains (level1_key, level2_key, level3_key)
        """
        result = []
        for level1_key, level1_dict in my_dict.items():
            for level2_key, level2_dict in level1_dict.items():
                for level3_key, value in level2_dict.items():
                    result.append(((level1_key, level2_key, level3_key), value))
        return result


    def tour_nested_dict_recursive(self, obj, current_keys=()):
        """
        Recursive function to tour through arbitrarily nested dictionaries.
        
        Args:
            obj: Dictionary or value to traverse
            current_keys: Current key path (used internally)
        
        Yields:
            tuple: (keys_tuple, value) where keys_tuple contains all keys in the path
        """
        if isinstance(obj, dict):
            for key, value in obj.items():
                yield from self.tour_nested_dict_recursive(value, current_keys + (key,))
        else:
            yield current_keys, obj
            
    @classmethod
    def pa_sky_to_image(cls, x, y, pa_east_of_north, wcs, offset_deg=0.001):
        """
        Convert Position Angle from astronomical standard (East of North) 
        to image coordinates (Right of Top) for plotting.
        
        Parameters
        ----------
        x : float
            X pixel coordinate (column)
        y : float
            Y pixel coordinate (row)
        pa_east_of_north : float
            Position angle in degrees, measured East of North
            (astronomical standard: 0° = North, 90° = East)
        wcs : astropy.wcs.WCS
            WCS object for coordinate transformation
        offset_deg : float, optional
            Small offset in degrees for numerical calculation (default: 0.001)
        
        Returns
        -------
        pa_image : float
            Position angle in image coordinates (degrees)
            Measured Right of Top (0° = up, 90° = right)
            This is the angle for matplotlib plotting
        """
        # Get sky coordinate at the reference point
        sky_center = wcs.pixel_to_world(x, y)
        
        # Calculate offset point in sky coordinates along the PA direction
        pa_rad = np.radians(pa_east_of_north)
        
        # East of North: 0° = North, 90° = East
        # dRA increases toward East, dDec increases toward North
        dra = offset_deg * np.sin(pa_rad) / np.cos(np.radians(sky_center.dec.deg))
        ddec = offset_deg * np.cos(pa_rad)
        
        # Create offset sky coordinate
        sky_offset = SkyCoord(
            ra=sky_center.ra + dra * u.deg,
            dec=sky_center.dec + ddec * u.deg,
            frame=sky_center.frame
        )
        
        # Transform both points to pixel coordinates
        x_center, y_center = wcs.world_to_pixel(sky_center)
        x_offset, y_offset = wcs.world_to_pixel(sky_offset)
        
        # Calculate image PA (Right of Top)
        # In image coordinates: +X is right, +Y is up (in standard matplotlib display)
        dx = x_offset - x_center
        dy = y_offset - y_center
        
        # arctan2(dx, dy) gives angle Right of Top
        pa_image = np.degrees(np.arctan2(dx, dy))
        
        return pa_image

    @classmethod
    def plot_pa_on_image(cls, image, x, y, pa_east_of_north, wcs, length=50, 
                        ax=None, color='red', linewidth=2, label=None):
        """
        Plot a position angle line on an astronomical image.
        
        Parameters
        ----------
        image : 2D array
            Image data to display
        x : float
            X pixel coordinate (center of PA line)
        y : float
            Y pixel coordinate (center of PA line)
        pa_east_of_north : float
            Position angle in degrees (East of North)
        wcs : astropy.wcs.WCS
            WCS object for the image
        length : float, optional
            Length of PA line in pixels (default: 50)
        ax : matplotlib.axes.Axes, optional
            Axes to plot on (creates new figure if None)
        color : str, optional
            Color of PA line (default: 'red')
        linewidth : float, optional
            Width of PA line (default: 2)
        label : str, optional
            Label for the PA line
        
        Returns
        -------
        ax : matplotlib.axes.Axes
            The axes object with the plot
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 10))
        
        # Display the image
        ax.imshow(image, origin='lower', cmap='gray', interpolation='nearest')
        
        # Convert PA to image coordinates
        pa_image = cls.pa_sky_to_image(x, y, pa_east_of_north, wcs)
        
        # Calculate line endpoints in image coordinates
        pa_rad = np.radians(pa_image)
        dx = length/2 * np.sin(pa_rad)
        dy = length/2 * np.cos(pa_rad)
        
        x1, y1 = x - dx, y - dy
        x2, y2 = x + dx, y + dy
        
        # Plot the PA line
        ax.plot([x1, x2], [y1, y2], color=color, linewidth=linewidth, 
                label=label if label else f'PA = {pa_east_of_north:.1f}° E of N')
        
        # Mark the center point
        ax.plot(x, y, 'o', color=color, markersize=8)
        
        # Add arrow at the positive direction
        ax.annotate('', xy=(x2, y2), xytext=(x, y),
                    arrowprops=dict(arrowstyle='->', color=color, lw=linewidth))
        
        return ax

    @classmethod
    def plot_compass_rose(cls, ax, x, y, wcs, size=30, color='cyan'):
        """
        Add a compass rose showing North and East directions on the image.
        
        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes to plot on
        x : float
            X pixel coordinate for compass center
        y : float
            Y pixel coordinate for compass center
        wcs : astropy.wcs.WCS
            WCS object
        size : float, optional
            Size of compass arrows in pixels (default: 30)
        color : str, optional
            Color of compass arrows (default: 'cyan')
        """
        # North direction (PA = 0°)
        pa_north = cls.pa_sky_to_image(x, y, 0, wcs)
        dx_n = size * np.sin(np.radians(pa_north))
        dy_n = size * np.cos(np.radians(pa_north))
        
        # East direction (PA = 90°)
        pa_east = cls.pa_sky_to_image(x, y, 90, wcs)
        dx_e = size * np.sin(np.radians(pa_east))
        dy_e = size * np.cos(np.radians(pa_east))
        
        # Plot North arrow
        ax.annotate('',xy=(x + dx_n, y + dy_n), xytext=(x, y),
                    arrowprops=dict(
                    arrowstyle='-|>',
                    lw=2,
                    color=color,
                    mutation_scale=5,
                    shrinkA=0,
                    shrinkB=0))
        ax.text(x + dx_n*1.2, y + dy_n*1.2, 'N', color=color, 
                fontsize=7, fontweight='bold', ha='center', va='center')
        
        # Plot East arrow
        ax.annotate('',xy=(x + dx_e, y + dy_e), xytext=(x, y),
                    arrowprops=dict(
                    arrowstyle='-|>',
                    lw=2,
                    color=color,
                    mutation_scale=5,
                    shrinkA=0,
                    shrinkB=0))
        ax.text(x + dx_e*1.2, y + dy_e*1.2, 'E', color=color,
                fontsize=7, fontweight='bold', ha='center', va='center')
        
    @classmethod
    def plot_scale(cls, ax, x, y, wcs, size=30, color='cyan'):
        """
        Add a compass rose showing North and East directions on the image.
        
        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes to plot on
        x : float
            X pixel coordinate for compass center
        y : float
            Y pixel coordinate for compass center
        wcs : astropy.wcs.WCS
            WCS object
        size : float, optional
            Size of compass arrows in arcseconds (default: 30)
        color : str, optional
            Color of compass arrows (default: 'cyan')
        """
        
        pixel_scale = cls.get_pixel_scale(wcs.to_header())
        size_pix = size / pixel_scale
        
        if size < 60:
            scale_text =  f'{size}'+r'$^{\prime\prime}$'
        elif (size >= 60) and (size < 3600):
            scale_text =  f'{size/60:.1f}'+r'$^\prime$'
        else:
            scale_text =  f'{size/3600:.1f}'+r'$^\circ$'
        
        ax.annotate('',xy=(x + size_pix/2, y), xytext=(x - size_pix/2, y),
                    arrowprops=dict(
                    arrowstyle='-',
                    lw=2,
                    color=color,
                    mutation_scale=5,
                    shrinkA=0,
                    shrinkB=0))
        ax.text(x, y + size_pix * 0.3, scale_text, color=color,
                fontsize=7, fontweight='bold', ha='center', va='center')


import matplotlib.pyplot as plt
import matplotlib.axes
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
from astropy.visualization import (
    AsinhStretch, AsymmetricPercentileInterval,
    ImageNormalize,
)
from astropy.stats import sigma_clipped_stats
from astropy.wcs import WCS
from scipy.stats import percentileofscore

# from ..utils.file_handler import GalaxyImageSet
from ..utils.utility import useful_functions


plt.rcParams["font.family"] = "FreeSerif"

plt.rcParams['axes.labelsize'] = 7
plt.rcParams['xtick.labelsize'] = 7
plt.rcParams['ytick.labelsize'] = 7


class DrawGalaxy:
    def __init__(self):
        pass
    
    @classmethod
    def plot_galaxies(cls, image_set, galaxy: str, step: str):
        image_dict = {}
        
        galaxy_data = image_set.data[galaxy]
        
        for obs, obs_dict in galaxy_data.items():
            for band, band_im in obs_dict.items():
                image_dict[f"{obs}.{band}"] = band_im
        
             
        m, n = useful_functions.find_rec(len(image_dict))
        
        fig, axes = plt.subplots(m, n, dpi=200, figsize=(n * 1.5, m * 1.2))
        
        for ax, im_data in zip(axes.flatten(), image_dict.items()):
            norm = ImageNormalize(im_data[1], interval=AsymmetricPercentileInterval(50., 99.8), stretch=AsinhStretch())
            
            im = ax.imshow(im_data[1], cmap='gray', origin="lower", norm=norm)
            
            ax.tick_params(axis="both", which="both", direction="in")
            ax.tick_params(axis="both", which="major", width=1.2)
            
            divider = make_axes_locatable(ax)
            cax = divider.append_axes('right', size='5%', pad=0.05)
            fig.colorbar(im, cax=cax, orientation='vertical')
            
        for ax in axes.flatten()[len(image_dict):]:
            ax.remove()
        
        fig.suptitle(f"Step Name: {step}")
        fig.tight_layout()
        plt.show()
    
    @classmethod
    def single_galaxy(cls, image_set, galaxy: str, obs: str, band: str, compass=True, scale=True):
        im_data = image_set.data[galaxy][obs][band]
        im_hdr = image_set.header[galaxy][obs][band]
        x, y = im_data.shape
        
        fig, ax = plt.subplots(1, 1, dpi=200, figsize=(4, 4), subplot_kw=dict(projection=WCS(im_hdr)))
            
        _, median, _ = sigma_clipped_stats(im_data, sigma=10.0, mask=np.where(im_data == 0, True, False))
        bg_percent = percentileofscore(np.nan_to_num(im_data.flatten()), median, kind="mean")
        
        norm = ImageNormalize(im_data, interval=AsymmetricPercentileInterval(bg_percent, 99.8), stretch=AsinhStretch())
        
        im = ax.imshow(im_data, cmap='gray', origin="lower", norm=norm)
        
        if compass:
            useful_functions.plot_compass_rose(ax, x * 0.9, y * 0.1, WCS(im_hdr), size=15, color='white')
            
        if scale:
            useful_functions.plot_scale(ax, x * 0.1, y * 0.1, WCS(im_hdr), size=120, color='white')
        
        ax.tick_params(axis="both", which="both", direction="in", color="#CCC")
        ax.tick_params(axis="both", which="major", width=1.2)
        ax.set_xlabel(r"$\alpha_{2000}$")
        ax.set_ylabel(r"$\delta_{2000}$")
        
        divider = make_axes_locatable(ax)
        cax = divider.append_axes('right', size='5%', pad=0.05, axes_class=matplotlib.axes.Axes)
        fig.colorbar(im, cax=cax, orientation='vertical')
        
        ax.set_title(f"{galaxy} {obs} {band}")
        fig.tight_layout()
        
        plt.show()
        
        return fig, ax